from threading import Thread
import rich_click as click
from rich import print
from rich.syntax import Syntax
from rich.prompt import Prompt
from rich.console import Console

from .utils import (
    TimeDeltaParamType,
    LiveData,
    background_progress_viewer,
    background_printer,
)
from .discovery.main import type_discovery
from .data import subscribe as data_subscribe


@click.command(short_help="Subscribe to an arbitrary topic")
@click.argument("topic")
@click.option(
    "-i", "--id", "--domain-id", type=int, default=0, help="DDS Domain to inspect."
)
@click.option(
    "-r",
    "--runtime",
    type=TimeDeltaParamType(),
    default="1s",
    help="Duration of discovery scan.",
)
@click.option(
    "--suppress-progress-bar",
    type=bool,
    is_flag=True,
    help="Suppress the output of the progress bar",
)
@click.option(
    "--color",
    type=click.Choice(["auto", "standard", "256", "truecolor", "windows", "none"]),
    default="auto",
    help="""Force the command to output with/without terminal colors. By default output colours if the terminal supports it."
See the [underline blue][link=https://rich.readthedocs.io/en/stable/console.html#color-systems]Rich documentation[/link][/] for more info on what the options mean.""",
)
def subscribe(topic, id, runtime, suppress_progress_bar, color):
    """Subscribe to an arbitrary topic"""
    console = Console(color_system=None if color == "none" else color)
    live = LiveData(console)

    thread = Thread(target=type_discovery, args=(live, id, runtime, topic))
    thread.start()

    console.print()
    background_progress_viewer(runtime, live, suppress_progress_bar)

    thread.join()

    if not live.result:
        console.print(
            "[bold red] :police_car_light: No types could be discovered over XTypes, no dynamic subsciption possible[/]"
        )
        return
    elif len(live.result) > 1:
        console.print(
            "[bold orange] :warning: Multiple type definitions exist, please pick one"
        )

        for i, (_, code, pp) in enumerate(live.result):
            console.print(
                f"Type {i}, As defined in participant(s) [magenta]"
                + ", ".join(f"{p}" for p in pp)
            )
            console.print(Syntax(code, "omg-idl"))
            console.print()

        index = Prompt.ask(
            "Please pick a type:",
            choices=[f"Type {i}" for i in range(len(live.result))],
        )
        index = int(index[len("Type: ") :])

        datatype = live.result[i][0]
    else:
        datatype = live.result[0][0]

    console.print("[bold green] Subscribing, CTRL-C to quit")

    thread = Thread(target=data_subscribe, args=(live, id, topic, datatype))
    thread.start()

    console.print()
    background_printer(live)
    console.print()

    thread.join()



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from . import internal, util, qos, core, domain, topic, pub, sub, builtin, dynamic, idl

__all__ = [
    "internal",
    "util",
    "qos",
    "core",
    "domain",
    "topic",
    "pub",
    "sub",
    "builtin",
    "dynamic",
    "idl",
]



import time
from typing import Any
from cyclonedds import domain, sub, topic

from ..utils import LiveData


def subscribe(live: LiveData, domain_id: int, topic_name: str, datatype: Any):
    dp = domain.DomainParticipant(domain_id)
    tp = topic.Topic(dp, topic_name, datatype)
    rd = sub.DataReader(dp, tp)

    while not live.terminate:
        for data in rd.take(N=20):
            live.printables.put(data, block=True)

        # yield thread
        time.sleep(0.01)

    live.delivered = True



"""
  Generated by Eclipse Cyclone DDS idlc Python Backend
  Cyclone DDS IDL version: v0.9.0
  Module: DDS.XTypes

"""

from ._ddsi_xt_type_object import ANNOTATION_OCTETSEC_VALUE_MAX_LEN, ANNOTATION_STR_VALUE_MAX_LEN, AliasMemberFlag, AliasTypeFlag, AnnotationParameterFlag, AnnotationParameterValue, AnnotationTypeFlag, AppliedAnnotation, AppliedAnnotationParameter, AppliedAnnotationParameterSeq, AppliedAnnotationSeq, AppliedBuiltinMemberAnnotations, AppliedBuiltinTypeAnnotations, AppliedVerbatimAnnotation, BitBound, BitflagFlag, BitmaskTypeFlag, BitsetMemberFlag, BitsetTypeFlag, CollectionElementFlag, CollectionTypeFlag, CommonAliasBody, CommonAnnotationParameter, CommonArrayHeader, CommonBitfield, CommonBitflag, CommonBitmaskHeader, CommonCollectionElement, CommonCollectionHeader, CommonDiscriminatorMember, CommonEnumeratedHeader, CommonEnumeratedLiteral, CommonStructMember, CommonUnionMember, CompleteAliasBody, CompleteAliasHeader, CompleteAliasType, CompleteAnnotationHeader, CompleteAnnotationParameter, CompleteAnnotationParameterSeq, CompleteAnnotationType, CompleteArrayHeader, CompleteArrayType, CompleteBitfield, CompleteBitfieldSeq, CompleteBitflag, CompleteBitflagSeq, CompleteBitmaskHeader, CompleteBitmaskType, CompleteBitsetHeader, CompleteBitsetType, CompleteCollectionElement, CompleteCollectionHeader, CompleteDiscriminatorMember, CompleteElementDetail, CompleteEnumeratedHeader, CompleteEnumeratedLiteral, CompleteEnumeratedLiteralSeq, CompleteEnumeratedType, CompleteExtendedType, CompleteMapType, CompleteMemberDetail, CompleteSequenceType, CompleteStructHeader, CompleteStructMember, CompleteStructMemberSeq, CompleteStructType, CompleteTypeDetail, CompleteTypeObject, CompleteUnionHeader, CompleteUnionMember, CompleteUnionMemberSeq, CompleteUnionType, EK_BOTH, EK_COMPLETE, EK_MINIMAL, EnumTypeFlag, EnumeratedLiteralFlag, EquivalenceHash, EquivalenceKind, ExtendedAnnotationParameterValue, ExtendedTypeDefn, INVALID_LBOUND, INVALID_SBOUND, LBound, LBoundSeq, MEMBER_NAME_MAX_LENGTH, MemberFlag, MemberFlagMinimalMask, MemberId, MemberName, MinimalAliasBody, MinimalAliasHeader, MinimalAliasType, MinimalAnnotationHeader, MinimalAnnotationParameter, MinimalAnnotationParameterSeq, MinimalAnnotationType, MinimalArrayHeader, MinimalArrayType, MinimalBitfield, MinimalBitfieldSeq, MinimalBitflag, MinimalBitflagSeq, MinimalBitmaskHeader, MinimalBitmaskType, MinimalBitsetHeader, MinimalBitsetType, MinimalCollectionElement, MinimalCollectionHeader, MinimalDiscriminatorMember, MinimalEnumeratedHeader, MinimalEnumeratedLiteral, MinimalEnumeratedLiteralSeq, MinimalEnumeratedType, MinimalExtendedType, MinimalMapType, MinimalMemberDetail, MinimalSequenceType, MinimalStructHeader, MinimalStructMember, MinimalStructMemberSeq, MinimalStructType, MinimalTypeDetail, MinimalTypeObject, MinimalUnionHeader, MinimalUnionMember, MinimalUnionMemberSeq, MinimalUnionType, NameHash, PlainArrayLElemDefn, PlainArraySElemDefn, PlainCollectionHeader, PlainMapLTypeDefn, PlainMapSTypeDefn, PlainSequenceLElemDefn, PlainSequenceSElemDefn, PrimitiveTypeId, QualifiedTypeName, SBound, SBoundSeq, StringLTypeDefn, StringSTypeDefn, StronglyConnectedComponent, StronglyConnectedComponentId, StructMemberFlag, StructTypeFlag, TI_PLAIN_ARRAY_LARGE, TI_PLAIN_ARRAY_SMALL, TI_PLAIN_MAP_LARGE, TI_PLAIN_MAP_SMALL, TI_PLAIN_SEQUENCE_LARGE, TI_PLAIN_SEQUENCE_SMALL, TI_STRING16_LARGE, TI_STRING16_SMALL, TI_STRING8_LARGE, TI_STRING8_SMALL, TI_STRONGLY_CONNECTED_COMPONENT, TK_ALIAS, TK_ANNOTATION, TK_ARRAY, TK_BITMASK, TK_BITSET, TK_BOOLEAN, TK_BYTE, TK_CHAR16, TK_CHAR8, TK_ENUM, TK_FLOAT128, TK_FLOAT32, TK_FLOAT64, TK_INT16, TK_INT32, TK_INT64, TK_MAP, TK_NONE, TK_SEQUENCE, TK_STRING16, TK_STRING8, TK_STRUCTURE, TK_UINT16, TK_UINT32, TK_UINT64, TK_UNION, TYPE_NAME_MAX_LENGTH, TypeFlag, TypeFlagMinimalMask, TypeIdentfierWithSizeSeq, TypeIdentiferKind, TypeIdentifier, TypeIdentifierPair, TypeIdentifierPairSeq, TypeIdentifierSeq, TypeIdentifierTypeObjectPair, TypeIdentifierTypeObjectPairSeq, TypeIdentifierWithDependencies, TypeIdentifierWithDependenciesSeq, TypeIdentifierWithSize, TypeInformation, TypeInformationSeq, TypeKind, TypeObject, TypeObjectHashId, TypeObjectSeq, UnionCaseLabelSeq, UnionDiscriminatorFlag, UnionMemberFlag, UnionTypeFlag
from ._ddsi_xt_typemap import TypeMapping
__all__ = ["ANNOTATION_OCTETSEC_VALUE_MAX_LEN", "ANNOTATION_STR_VALUE_MAX_LEN", "AliasMemberFlag", "AliasTypeFlag", "AnnotationParameterFlag", "AnnotationParameterValue", "AnnotationTypeFlag", "AppliedAnnotation", "AppliedAnnotationParameter", "AppliedAnnotationParameterSeq", "AppliedAnnotationSeq", "AppliedBuiltinMemberAnnotations", "AppliedBuiltinTypeAnnotations", "AppliedVerbatimAnnotation", "BitBound", "BitflagFlag", "BitmaskTypeFlag", "BitsetMemberFlag", "BitsetTypeFlag", "CollectionElementFlag", "CollectionTypeFlag", "CommonAliasBody", "CommonAnnotationParameter", "CommonArrayHeader", "CommonBitfield", "CommonBitflag", "CommonBitmaskHeader", "CommonCollectionElement", "CommonCollectionHeader", "CommonDiscriminatorMember", "CommonEnumeratedHeader", "CommonEnumeratedLiteral", "CommonStructMember", "CommonUnionMember", "CompleteAliasBody", "CompleteAliasHeader", "CompleteAliasType", "CompleteAnnotationHeader", "CompleteAnnotationParameter", "CompleteAnnotationParameterSeq", "CompleteAnnotationType", "CompleteArrayHeader", "CompleteArrayType", "CompleteBitfield", "CompleteBitfieldSeq", "CompleteBitflag", "CompleteBitflagSeq", "CompleteBitmaskHeader", "CompleteBitmaskType", "CompleteBitsetHeader", "CompleteBitsetType", "CompleteCollectionElement", "CompleteCollectionHeader", "CompleteDiscriminatorMember", "CompleteElementDetail", "CompleteEnumeratedHeader", "CompleteEnumeratedLiteral", "CompleteEnumeratedLiteralSeq", "CompleteEnumeratedType", "CompleteExtendedType", "CompleteMapType", "CompleteMemberDetail", "CompleteSequenceType", "CompleteStructHeader", "CompleteStructMember", "CompleteStructMemberSeq", "CompleteStructType", "CompleteTypeDetail", "CompleteTypeObject", "CompleteUnionHeader", "CompleteUnionMember", "CompleteUnionMemberSeq", "CompleteUnionType", "EK_BOTH", "EK_COMPLETE", "EK_MINIMAL", "EnumTypeFlag", "EnumeratedLiteralFlag", "EquivalenceHash", "EquivalenceKind", "ExtendedAnnotationParameterValue", "ExtendedTypeDefn", "INVALID_LBOUND", "INVALID_SBOUND", "LBound", "LBoundSeq", "MEMBER_NAME_MAX_LENGTH", "MemberFlag", "MemberFlagMinimalMask", "MemberId", "MemberName", "MinimalAliasBody", "MinimalAliasHeader", "MinimalAliasType", "MinimalAnnotationHeader", "MinimalAnnotationParameter", "MinimalAnnotationParameterSeq", "MinimalAnnotationType", "MinimalArrayHeader", "MinimalArrayType", "MinimalBitfield", "MinimalBitfieldSeq", "MinimalBitflag", "MinimalBitflagSeq", "MinimalBitmaskHeader", "MinimalBitmaskType", "MinimalBitsetHeader", "MinimalBitsetType", "MinimalCollectionElement", "MinimalCollectionHeader", "MinimalDiscriminatorMember", "MinimalEnumeratedHeader", "MinimalEnumeratedLiteral", "MinimalEnumeratedLiteralSeq", "MinimalEnumeratedType", "MinimalExtendedType", "MinimalMapType", "MinimalMemberDetail", "MinimalSequenceType", "MinimalStructHeader", "MinimalStructMember", "MinimalStructMemberSeq", "MinimalStructType", "MinimalTypeDetail", "MinimalTypeObject", "MinimalUnionHeader", "MinimalUnionMember", "MinimalUnionMemberSeq", "MinimalUnionType", "NameHash", "PlainArrayLElemDefn", "PlainArraySElemDefn", "PlainCollectionHeader", "PlainMapLTypeDefn", "PlainMapSTypeDefn", "PlainSequenceLElemDefn", "PlainSequenceSElemDefn", "PrimitiveTypeId", "QualifiedTypeName", "SBound", "SBoundSeq", "StringLTypeDefn", "StringSTypeDefn", "StronglyConnectedComponent", "StronglyConnectedComponentId", "StructMemberFlag", "StructTypeFlag", "TI_PLAIN_ARRAY_LARGE", "TI_PLAIN_ARRAY_SMALL", "TI_PLAIN_MAP_LARGE", "TI_PLAIN_MAP_SMALL", "TI_PLAIN_SEQUENCE_LARGE", "TI_PLAIN_SEQUENCE_SMALL", "TI_STRING16_LARGE", "TI_STRING16_SMALL", "TI_STRING8_LARGE", "TI_STRING8_SMALL", "TI_STRONGLY_CONNECTED_COMPONENT", "TK_ALIAS", "TK_ANNOTATION", "TK_ARRAY", "TK_BITMASK", "TK_BITSET", "TK_BOOLEAN", "TK_BYTE", "TK_CHAR16", "TK_CHAR8", "TK_ENUM", "TK_FLOAT128", "TK_FLOAT32", "TK_FLOAT64", "TK_INT16", "TK_INT32", "TK_INT64", "TK_MAP", "TK_NONE", "TK_SEQUENCE", "TK_STRING16", "TK_STRING8", "TK_STRUCTURE", "TK_UINT16", "TK_UINT32", "TK_UINT64", "TK_UNION", "TYPE_NAME_MAX_LENGTH", "TypeFlag", "TypeFlagMinimalMask", "TypeIdentfierWithSizeSeq", "TypeIdentiferKind", "TypeIdentifier", "TypeIdentifierPair", "TypeIdentifierPairSeq", "TypeIdentifierSeq", "TypeIdentifierTypeObjectPair", "TypeIdentifierTypeObjectPairSeq", "TypeIdentifierWithDependencies", "TypeIdentifierWithDependenciesSeq", "TypeIdentifierWithSize", "TypeInformation", "TypeInformationSeq", "TypeKind", "TypeObject", "TypeObjectHashId", "TypeObjectSeq", "UnionCaseLabelSeq", "UnionDiscriminatorFlag", "UnionMemberFlag", "UnionTypeFlag", "TypeMapping", ]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from enum import Enum, IntFlag, auto
from inspect import isclass
from typing import Tuple, Type, Union

from . import IdlStruct, IdlUnion, IdlBitmask
from ._type_helper import get_origin, get_args
from ._type_normalize import get_idl_annotations, get_idl_field_annotations, get_extended_type_hints, WrapOpt
from ._machinery import Machine, NoneMachine, PrimitiveMachine, StringMachine, BytesMachine, ByteArrayMachine, UnionMachine, \
    ArrayMachine, SequenceMachine, InstanceMachine, MappingMachine, EnumMachine, StructMachine, OptionalMachine, CharMachine, \
    PLCdrMutableStructMachine, DelimitedCdrAppendableStructMachine, MutableMember, DelimitedCdrAppendableUnionMachine, \
    PlainCdrV2ArrayOfPrimitiveMachine, PlainCdrV2SequenceOfPrimitiveMachine, LenType, BitMaskMachine, BitBoundEnumMachine

from .types import array, bounded_str, sequence, _type_code_align_size_default_mapping, NoneType, char, typedef, uint8, \
    case, default


class XCDRSupported(IntFlag):
    SupportsBasic = 1
    SupportsV2 = 2
    SupportsBoth = SupportsBasic | SupportsV2


class Builder:
    easy_types = {
        char: CharMachine,
        str: StringMachine,
        bytes: BytesMachine,
        bytearray: ByteArrayMachine,
        NoneType: NoneMachine,
        None: NoneMachine
    }

    @classmethod
    def _scan_for_support(cls, _type, done) -> XCDRSupported:
        if id(_type) in done:
            return XCDRSupported.SupportsBoth
        # to avoid infinite recursing. Since this is the 'all' flag it
        # will not affect the result, because of the tree structure it is
        # guaranteed to be binary and'ed with the real result.
        done.add(id(_type))

        if isinstance(_type, WrapOpt):
            return XCDRSupported.SupportsV2
        elif isinstance(_type, (typedef, sequence, array)):
            return cls._scan_for_support(_type.subtype, done)
        elif get_origin(_type) == list:
            return cls._scan_for_support(get_args(_type)[0], done)
        elif get_origin(_type) == dict:
            return cls._scan_for_support(get_args(_type)[0], done) & cls._scan_for_support(get_args(_type)[1], done)
        elif isclass(_type) and issubclass(_type, IdlStruct):
            fields = get_extended_type_hints(_type)
            annotations = get_idl_annotations(_type)

            # Explicit setter
            if 'xcdrv2' in annotations:
                if annotations['xcdrv2']:
                    return XCDRSupported.SupportsV2
                else:
                    return XCDRSupported.SupportsBasic

            # Appendable and mutable is definitely V2
            if 'extensibility' in annotations:
                if annotations['extensibility'] in ['appendable', 'mutable']:
                    return XCDRSupported.SupportsV2

            # Check for optionals or nested mutable/appendable
            support = XCDRSupported.SupportsBoth
            for _, _ftype in fields.items():
                support &= cls._scan_for_support(_ftype, done)

            return support
        elif isclass(_type) and issubclass(_type, IdlUnion):
            fields = get_extended_type_hints(_type)
            annotations = get_idl_annotations(_type)

            # Explicit setter
            if 'xcdrv2' in annotations:
                if annotations['xcdrv2']:
                    return XCDRSupported.SupportsV2
                else:
                    return XCDRSupported.SupportsBasic

            # Appendable and mutable is definitely V2
            if 'extensibility' in annotations:
                if annotations['extensibility'] in ['appendable', 'mutable']:
                    return XCDRSupported.SupportsV2

            # Check for optionals or nested mutable/appendable
            support = XCDRSupported.SupportsBoth
            for _, _ftype in fields.items():
                support &= cls._scan_for_support(_ftype.subtype, done)

            return support
        return XCDRSupported.SupportsBoth


    @classmethod
    def _machine_for_type(cls, _type, add_size_header, use_version_2):
        if _type in cls.easy_types:
            return cls.easy_types[_type]()
        elif _type in _type_code_align_size_default_mapping:
            return PrimitiveMachine(_type)
        elif isinstance(_type, WrapOpt):
            return OptionalMachine(cls._machine_for_type(_type.inner, add_size_header, use_version_2))
        elif isclass(_type) and issubclass(_type, Enum):
            if "bit_bound" in get_idl_annotations(_type) and use_version_2:
                return BitBoundEnumMachine(_type, get_idl_annotations(_type)["bit_bound"])
            return EnumMachine(_type)
        elif isclass(_type) and (issubclass(_type, IdlStruct) or issubclass(_type, IdlUnion)):
            return InstanceMachine(_type, use_version_2)
        elif isclass(_type) and (issubclass(_type, IdlBitmask)):
            return BitMaskMachine(_type, get_idl_annotations(_type)["bit_bound"])
        elif get_origin(_type) == list:
            return SequenceMachine(
                cls._machine_for_type(get_args(_type)[0], add_size_header, use_version_2),
                add_size_header=add_size_header
            )
        elif get_origin(_type) == dict:
            return MappingMachine(
                cls._machine_for_type(get_args(_type)[0], add_size_header, use_version_2),
                cls._machine_for_type(get_args(_type)[1], add_size_header, use_version_2)
            )
        elif isinstance(_type, typedef):
            return cls._machine_for_type(_type.subtype, add_size_header, use_version_2)
        elif isinstance(_type, array):
            submachine = cls._machine_for_type(_type.subtype, add_size_header, use_version_2)

            if isinstance(submachine, PrimitiveMachine):
                if submachine.type == uint8:
                    return ByteArrayMachine(_type.length)

                return PlainCdrV2ArrayOfPrimitiveMachine(submachine.type, _type.length)

            asubmachine = submachine
            while isinstance(asubmachine, ArrayMachine):
                asubmachine.add_size_header = False
                asubmachine = asubmachine.submachine

            if isinstance(asubmachine, (ByteArrayMachine, PlainCdrV2ArrayOfPrimitiveMachine, CharMachine)):
                add_size_header = False

            return ArrayMachine(
                submachine,
                size=_type.length,
                add_size_header=add_size_header
            )
        elif isinstance(_type, sequence):
            submachine = cls._machine_for_type(_type.subtype, add_size_header, use_version_2)

            if isinstance(submachine, PrimitiveMachine):
                return PlainCdrV2SequenceOfPrimitiveMachine(submachine.type, max_length=_type.max_length)

            if isinstance(submachine, (CharMachine)):
                add_size_header = False

            return SequenceMachine(
                submachine,
                maxlen=_type.max_length,
                add_size_header=add_size_header
            )
        elif isinstance(_type, bounded_str):
            return StringMachine(
                bound=_type.max_length
            )

        raise TypeError(f"{_type} is not valid in IDL classes because it cannot be encoded.")

    @classmethod
    def _machine_struct(cls, struct: Type[IdlStruct]) -> Tuple[Machine, bool]:
        fields = get_extended_type_hints(struct)
        annotations = get_idl_annotations(struct)
        field_annotations = get_idl_field_annotations(struct)

        keylist = annotations.get("keylist")
        if not keylist:
            keylist = []
            for name, f_annotations in field_annotations.items():
                if "key" in f_annotations:
                    keylist.append(name)

            if not keylist:
                keylist = None

        keyless = keylist is None or len(keylist) == 0

        extensibility = annotations.get("extensibility")

        v0_members = {
            name: cls._machine_for_type(field_type, False, False)
            for name, field_type in fields.items()
        }
        v2_members = {
            name: cls._machine_for_type(field_type, True, True)
            for name, field_type in fields.items()
        }

        v0_machine = StructMachine(struct, v0_members, keylist)

        if extensibility and extensibility != "final":
            if extensibility == "appendable":
                v2_machine = DelimitedCdrAppendableStructMachine(
                    struct, v2_members, keylist
                )
            elif extensibility == "mutable":
                mutablemembers = []

                for name, machine in v2_members.items():
                    optional = False
                    if isinstance(machine, OptionalMachine):
                        optional = True
                        machine = machine.submachine

                    lentype = None
                    if isinstance(machine, PrimitiveMachine):
                        lentype = {
                            1: LenType.OneByte,
                            2: LenType.TwoByte,
                            4: LenType.FourByte,
                            8: LenType.EightByte
                        }[machine.size]
                    elif isinstance(machine, PlainCdrV2SequenceOfPrimitiveMachine):
                        if machine.size == 1:
                            lentype = LenType.NextIntDualUseLen
                        elif machine.size == 4:
                            lentype = LenType.NextIntDualUse4Len
                        elif machine.size == 8:
                            lentype = LenType.NextIntDualUse8Len
                        else:
                            lentype = LenType.NextIntLen
                    elif isinstance(machine, SequenceMachine):
                        lentype = LenType.NextIntDualUseLen
                    else:
                        lentype = LenType.NextIntLen

                    mutablemembers.append(MutableMember(
                        name=name,
                        key=not keylist or name in keylist,
                        optional=optional,
                        lentype=lentype,
                        must_understand=field_annotations.get(name, {}).get("must_understand", False),
                        memberid=struct.__idl__.get_member_id(name),
                        machine=machine
                    ))

                v2_machine = PLCdrMutableStructMachine(struct, mutablemembers)
        else:
            v2_machine = StructMachine(struct, v2_members, keylist)

        return v0_machine, v2_machine, keyless

    @classmethod
    def _machine_union(cls, union: Type[IdlUnion]):
        annotations = get_idl_annotations(union)
        extensibility = annotations.get("extensibility")

        v0_cases = {}
        v2_cases = {}
        v0_default = None
        v2_default = None

        for _case in get_extended_type_hints(union).values():
            if isinstance(_case, default):
                v0_default = cls._machine_for_type(_case.subtype, False, False)
                v2_default = cls._machine_for_type(_case.subtype, True, True)
                continue

            if not isinstance(_case, case):
                continue

            v0_machine = cls._machine_for_type(_case.subtype, False, False)
            v2_machine = cls._machine_for_type(_case.subtype, True, True)

            for label in _case.labels:
                v0_cases[label] = v0_machine
                v2_cases[label] = v2_machine

        v0_discriminator = cls._machine_for_type(union.__idl_discriminator__, False, False)
        v2_discriminator = cls._machine_for_type(union.__idl_discriminator__, True, True)

        v0_machine = UnionMachine(union, v0_discriminator, v0_cases, v0_default)

        if extensibility is None or extensibility == "final":
            v2_machine = UnionMachine(union, v2_discriminator, v2_cases, v2_default)

        elif extensibility == "appendable":
            v2_machine = DelimitedCdrAppendableUnionMachine(union, v2_discriminator, v2_cases, v2_default)
        else:
            # mutable
            raise NotImplementedError()

        return v0_machine, v2_machine

    @classmethod
    def build_machines(cls, _type):
        if issubclass(_type, IdlUnion):
            v0_machine, v2_machine = cls._machine_union(_type)
            keyless = False
        elif issubclass(_type, IdlStruct):
            v0_machine, v2_machine, keyless = cls._machine_struct(_type)
        else:
            raise Exception(f"Cannot build for {_type}, not struct or union.")

        return v0_machine, v2_machine, keyless, cls._scan_for_support(_type, set())
